/*
 * File:	AnnoParser.java
 * author:	Haitao Liu (haitaol@cs.arizona.edu)
 *
 * The tauXSchema toolset, Version 1.0
 *
 * Copyright (c) 2002-2003 
 * The Department of Computer Science of the University of Arizona.
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by 
 *       the Department of Computer Science of the University of Arizona."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear. 
 *
 * 4. The names "tauXSchema" or "TAU (Temporal Access for Users)" 
 *    must not be used to endorse or promote products derived from this 
 *    software without prior written permission. For written permission, 
 *    please contact tau@cs.arizona.edu.
 *
 * 5. Products derived from this software may not be called "tauXSchema",
 *    nor may "tauXSchema" or "TAU (Temporal Access for Users)" appear in 
 *    their name, without prior written permission of the Department of 
 *    Computer Science of the University of Arizona.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE Department of Computer Science of 
 * the University of Arizona OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This toolset contains useful tools to manage temporal XML document.
 * For more information about tXSchema and temporal XML document, 
 * visit the homepage of TAU Project <http://www.cs.arizona.edu/tau/>.
 */

import javax.xml.parsers.*;
import org.w3c.dom.*;
import org.apache.xpath.*;
import java.util.*;
import org.apache.xerces.dom3.*;
import org.apache.xerces.parsers.*;
import org.w3c.dom.ls.*;
import java.io.*;
import org.w3c.dom.traversal.*;
import org.apache.xerces.parsers.*;
import org.xml.sax.*;

/**
 * The AnnoParser class keeps information about the logical annotations and 
 * physical annotations and provide the information to temporal XML applications
 * on request.
 * @author  Haitao Liu
 * @version 1.0, May 2003 
 */
public class AnnoParser
{	
	/* store paths to timestampable elements and attributes */
	ArrayList ts_paths;
	
	/* store paths to value-varying elements */
	ArrayList vv_paths;
	
	/* store path to non-existence-varying elements */
	ArrayList ev_paths;
	
	/* store path to timevarying attributes */
	ArrayList tvattr_paths;
	
	/* logical and physical annotation documents */
	Document log_doc, phy_doc;

	/* 
	 *	Creates an AnnoParser to keep the information of the given logical and physical annotations.
	 *	@param   log_doc	parsed logical annotation document
	 *	@param   phy_doc	parsed physical annotation document
	 */	
	public AnnoParser(Document log_doc, Document phy_doc){
		this.log_doc = log_doc;
		this.phy_doc = phy_doc;
		
		ts_paths = new ArrayList();			
		vv_paths = new ArrayList();			
		ev_paths = new ArrayList();			
		tvattr_paths = new ArrayList();				
		
		if (log_doc != null){		
			/* get paths to time-varying elements and attributes in logical annotation */
			
			NodeList vt = log_doc.getElementsByTagName("validTime");				
			
			for (int i = 0; i < vt.getLength(); i++){
				Element vte = (Element)vt.item(i);
				String target = vte.getAttribute("target");
				
				if (target.indexOf("@") >= 0)
					tvattr_paths.add(target);
				else{
					if (vte.getAttribute("valueVarying").equals("true"))
						vv_paths.add(target);
					
					if (vte.getAttribute("existenceVarying").equals("true"))
						ev_paths.add(target);
				}
			}
		}
		
		if (phy_doc != null){
			/* get paths to timestampable elements and attributes in physical annotation */
			
			NodeList stamp_posi = phy_doc.getElementsByTagName("stampPosition");
			
			for (int i = 0; i < stamp_posi.getLength(); i++){			
				Element spe = (Element)stamp_posi.item(i);
				
				String target = spe.getAttribute("target");		
				
				ts_paths.add(target);
			}
		}
	}
	
	/**
	 *	Test if the given element is timevarying.
	 *	@param	e	The element to be Tested.
	 *	@return  true if the element is timevarying and false otherwise.
	 */
	public boolean isTimeVarying(Element e){
		return isValueVarying(e)|isExistenceVarying(e);		
	}
	
	/* get the complete path from the root to given element */
	String get_path(Element e){
		String path = e.getNodeName();
		
		Node parent = e.getParentNode();
		
		while (parent instanceof Element){
			if (parent.getNodeName().equals("timevarying-root"))
				break;
				
			path = parent.getNodeName() + "/" + path;			
				
			parent = parent.getParentNode();
		}
		
		path = "/" + path;
		
		return path;
	}
	
	/**
	 *	Test if the given element can have timestamp.
	 *	@param	e	The element to be Tested.
	 *	@return  true if the element can have timestamp and false otherwise.
	 */
	public boolean isTimestampableElement(Element e){
		String path = get_path(e);
		
		if (ts_paths.indexOf(path) >= 0)
			return true;
		else 
			return false;
	}
	
	/**
	 * Test if the given element is value-varying.
	 * @param	e	The element to be Tested.
	 * @return  true if the element is value-varying and false otherwise.
	 */
	public boolean isValueVarying(Element e){
		String path = get_path(e);
		
		if (vv_paths.indexOf(path) >= 0)
			return true;
		else 
			return false;
	}
	
	/**
	 * Test if the given element is existence-varying.
	 * @param	e	The element to be Tested.
	 * @return  true if the element is existence-varying and false otherwise.
	 */
	public boolean isExistenceVarying(Element e){
		String path = get_path(e);
		
		if (ev_paths.indexOf(path) >= 0)
			return true;
		else 
			return false;
	}
	
	/**
	 * Return all paths leading to elements that are value-varying.
	 * @return  An arraylist containing the paths.
	 */
	public ArrayList getValueVaryingElementsPath(){
		return (ArrayList)vv_paths.clone();
	}
	
	/**
	 * Return all paths leading to elements that are existence-varying.
	 * @return  An arraylist containing the paths.
	 */
	public ArrayList getExistenceVaryingElementsPath(){
		return (ArrayList)ev_paths.clone();
	}
	
	/**
	 * Return all paths leading to time-varying attributes.
	 * @return  An arraylist containing the paths.
	 */
	public ArrayList getTimeVaryingAttrPath(){
		return (ArrayList)tvattr_paths.clone();
	}
	
	/**
	 * Return all paths leading to elements that can have timestamp.
	 * @return  An arraylist containing the paths.
	 */
	public ArrayList getTimestampableElementsPath(){
		ArrayList path_list = new ArrayList();
		
		for (int i = 0; i < ts_paths.size(); i++){
			String path = (String)ts_paths.get(i);
			
			if (path.indexOf("@") < 0)
				path_list.add(path);
		}
		
		return path_list;
	}
	
	/**
	 * Return all paths leading to timevarying attributes.
	 * @return  An arraylist containing the paths.
	 */
	public ArrayList getTVAttrPath(){
		return (ArrayList)tvattr_paths.clone();
	}
	
	/*
	 * Return all paths leading to timevarying attributes that can have 
	 * timestamps.
	 * @return  An arraylist containing the paths.
	 */
	public ArrayList getTimestampableTVAttrPath(){
		ArrayList path_list = new ArrayList();
		
		for (int i = 0; i < ts_paths.size(); i++){
			String path = (String)ts_paths.get(i);
			
			if (path.indexOf("@") >= 0)
				path_list.add(path);
		}
		
		return path_list;
	}
	
	/**
	 * Test if the given element has a timevarying attribute with the given name.
	 * @param 	e		The given element.
	 * @param	name	The name of the timevarying attribute.
	 * @return  true if that element has timevarying attribute with that name and false otherwise.
	 */
	public boolean hasTVAttr(Element e, String name){
		String path = get_path(e) + "/@" + name;
		
		if (tvattr_paths.indexOf(path) >= 0)
			return true;
		else		
			return false;
	}
	
	/**
	 * Test if the given element has a timevarying attribute and that attribute is timestampable
	 * @param 	e		The given element.
	 * @param	name	The name of the timevarying attribute.
	 * @return  true if that element has timestampable timevarying attribute with that name and false otherwise.
	 */
	public boolean hasTimestampableTVAttr(Element e, String name){
		String path = get_path(e) + "/@" + name;
		
		if (ts_paths.indexOf(path) >= 0)
			return true;
		else 
			return false;
	}
}
