/*
 * File:	AnnoValidator.java
 * author:	Haitao Liu (haitaol@cs.arizona.edu)
 *
 * The tauXSchema toolset, Version 1.0
 *
 * Copyright (c) 2002-2003 
 * The Department of Computer Science of the University of Arizona.
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by 
 *       the Department of Computer Science of the University of Arizona."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear. 
 *
 * 4. The names "tauXSchema" or "TAU (Temporal Access for Users)" 
 *    must not be used to endorse or promote products derived from this 
 *    software without prior written permission. For written permission, 
 *    please contact tau@cs.arizona.edu.
 *
 * 5. Products derived from this software may not be called "tauXSchema",
 *    nor may "tauXSchema" or "TAU (Temporal Access for Users)" appear in 
 *    their name, without prior written permission of the Department of 
 *    Computer Science of the University of Arizona.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE Department of Computer Science of 
 * the University of Arizona OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This toolset contains useful tools to manage temporal XML document.
 * For more information about tXSchema and temporal XML document, 
 * visit the homepage of TAU Project <http://www.cs.arizona.edu/tau/>.
 */

import org.w3c.dom.*;
import java.util.*;
import org.w3c.dom.traversal.*;

/**
 * The AnnoValidator class test if the given logical and/or physical annotation are valid for
 * the specified non-temporal schema. 
 * @author  Haitao Liu
 * @version 1.0, May 2003 
 */
public class AnnoValidator
{
	static Hashtable def_table;
	
	static ArrayList vt_path, stamp_path;
			
	static void get_top_defs(Document schema){
		def_table = new Hashtable();
		
		NodeList top_child = schema.getDocumentElement().getChildNodes();
		
		for (int i = 0; i < top_child.getLength(); i++){
			Node n = top_child.item(i);
			
			if (n.getNodeName().equals("xs:element"))
				def_table.put(((Element)n).getAttribute("name"), n);	
		}
		
		return;		
	}
	
	/**
	 * Test both logical annotation and physical annotation at the same time. Beyond the checks
	 * in separate tests, a node denoted timevarying in logical annotation should have an ancestor 
	 * denoted timestampable in physical annotation.
	 * @param 	schema		Non-temporal schema document.
	 * @param	log_anno	Logical annotation document.
	 * @param	phy_anno	Physical annotation document.
	 * @return  true if annotations are valid and false otherwise.
	 */
	public static boolean validLogPhyAnno(Document schema, Document log_anno, Document phy_anno){
		if (!validLogicalAnno(schema, log_anno))			
			return false;
			
		if (!validPhysicalAnno(schema, phy_anno))			
			return false;
			
		for (int i = 0; i < vt_path.size(); i++){
			String vt_p = (String)vt_path.get(i);
			
			int j;
			for (j = 0; j < stamp_path.size(); j++){
				String stamp_p = (String)stamp_path.get(j);
				
				if (vt_p.startsWith(stamp_p))
					break;	
			}
			
			if (j >= stamp_path.size())
				return false;
		}
		
		return true;
	}

	/**
	 * Test the logical annotation. Paths indicating timevarying parts in logical annotation should 
	 * lead to valid entities in snapshots whose structure are described by non-temporal schema.
	 * @param 	schema		Non-temporal schema document.
	 * @param	log_anno	Logical annotation document.	 
	 * @return  true if annotation is valid and false otherwise.
	 */
	public static boolean validLogicalAnno(Document schema, Document log_anno){
		get_top_defs(schema);
		
		NodeList vt_list = log_anno.getElementsByTagName("validTime");
		
		vt_path = new ArrayList();
		
		for (int i = 0; i < vt_list.getLength(); i++){
			String target = ((Element)vt_list.item(i)).getAttribute("target");
			
			if (!testPath(target))
				return false;
				
			vt_path.add(target);
		}
		
		return true;
	}	
	
	/**
	 * Test the physical annotation. Paths indicating timestampable parts in physical annotation should 
	 * lead to valid entities in snapshots whose structure are described by non-temporal schema.
	 * @param 	schema		Non-temporal schema document.
	 * @param	phy_anno	physical annotation document.	 
	 * @return  true if annotation is valid and false otherwise.
	 */	
	public static boolean validPhysicalAnno(Document schema, Document phy_anno){
		get_top_defs(schema);
		
		NodeList stamppos_list = phy_anno.getElementsByTagName("stampPosition");
		
		stamp_path = new ArrayList();
		
		for (int i = 0; i < stamppos_list.getLength(); i++){
			String target = ((Element)stamppos_list.item(i)).getAttribute("target");
			
			if (!testPath(target))
				return false;
				
			stamp_path.add(target);
		}
		
		return true;
	}	
	
	/* test if a path given in annotation is a valid according to schema */
	static boolean testPath(String path){
		String attr_name = null;
		
		if (path.startsWith("/"))
			path = path.substring(1);
			
		if (path.indexOf("@") >= 0){
			attr_name = path.substring(path.indexOf("@") + 1);
			path = path.substring(0, path.indexOf("@") - 1);
		}
		
		String[] nodes_name	= path.split("/");
		
		Element parent_def = null;
		
		for (int i = 0; i < nodes_name.length; i++){
			String name = nodes_name[i];
			
			if (parent_def == null){
				parent_def = (Element)def_table.get(name);
				
				if (parent_def == null)				
					return false;
					
				continue;
			}
			
			TreeWalker tw_def = ((DocumentTraversal)parent_def.getOwnerDocument()).createTreeWalker(parent_def,
														NodeFilter.SHOW_ELEMENT,
														null,
														false);
			
			ArrayList past_nodes = new ArrayList();
			
			Element def = (Element)tw_def.firstChild();
			while (true){
				if (def == null){
					if (past_nodes.size() == 0)
						break;
						
					def = (Element)past_nodes.remove(past_nodes.size() - 1);
					
					tw_def.setCurrentNode(def);
					def = (Element)tw_def.nextSibling();
					
					continue;
				}			
				
				if (def.getNodeName().equals("xs:element")){
					if (def.getAttribute("name").equals(name)){
						parent_def = def;
						break;	
					}
					
					if (def.getAttribute("ref").equals(name)){
						parent_def = (Element)def_table.get(name);
						break;	
					}
					
					tw_def.setCurrentNode(def);
					def = (Element)tw_def.nextSibling();
				}	
				else{
					past_nodes.add(def);
					
					tw_def.setCurrentNode(def);
					def = (Element)tw_def.firstChild();
				}				
			}
			
			if (def == null)
				return false;
		}
		
		if (attr_name != null){
			Element owner = parent_def;
			
			TreeWalker tw_attr = ((DocumentTraversal)owner.getOwnerDocument()).createTreeWalker(owner,
														NodeFilter.SHOW_ELEMENT,
														null,
														false);
														
			ArrayList past_nodes = new ArrayList();
														
			Element def = (Element)tw_attr.firstChild();
			while (true){
				if (def == null){
					if (past_nodes.size() == 0)
						return false;
						
					def = (Element)past_nodes.remove(past_nodes.size() - 1);
					
					tw_attr.setCurrentNode(def);
					def = (Element)tw_attr.nextSibling();
					
					continue;
				}			
				
				if (def.getNodeName().equals("xs:attribute")){
					if (def.getAttribute("name").equals(attr_name))
						return true;
					
					if (def.getAttribute("ref").equals(attr_name))
						return true;
					
					tw_attr.setCurrentNode(def);
					def = (Element)tw_attr.nextSibling();
				}	
				else if (def.getNodeName().equals("xs:element")){
					tw_attr.setCurrentNode(def);
					def = (Element)tw_attr.nextSibling();
				}
				else{
					past_nodes.add(def);
					
					tw_attr.setCurrentNode(def);
					def = (Element)tw_attr.firstChild();
				}
			}
		}
		else
			return true;
	}
}
