/*
 * File:	SchemaMapper.java
 * author:	Haitao Liu (haitaol@cs.arizona.edu)
 *
 * The tauXSchema toolset, Version 1.0
 *
 * Copyright (c) 2002-2003 
 * The Department of Computer Science of the University of Arizona.
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by 
 *       the Department of Computer Science of the University of Arizona."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear. 
 *
 * 4. The names "tauXSchema" or "TAU (Temporal Access for Users)" 
 *    must not be used to endorse or promote products derived from this 
 *    software without prior written permission. For written permission, 
 *    please contact tau@cs.arizona.edu.
 *
 * 5. Products derived from this software may not be called "tauXSchema",
 *    nor may "tauXSchema" or "TAU (Temporal Access for Users)" appear in 
 *    their name, without prior written permission of the Department of 
 *    Computer Science of the University of Arizona.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE Department of Computer Science of 
 * the University of Arizona OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This toolset contains useful tools to manage temporal XML document.
 * For more information about tXSchema and temporal XML document, 
 * visit the homepage of TAU Project <http://www.cs.arizona.edu/tau/>.
 */

import javax.xml.parsers.*;
import org.w3c.dom.*;
import org.apache.xpath.*;
import java.util.*;
import org.apache.xerces.dom3.*;
import org.apache.xerces.parsers.*;
import org.w3c.dom.ls.*;
import java.io.*;
import org.w3c.dom.traversal.*;
import org.apache.xerces.parsers.*;
import org.xml.sax.*;

public class SchemaMapper
{
	static DOMParser parser;
	static DOMWriter writer;	
	
	static String schema_name;
	static Document schema = null, phy_anno = null;
	static Element sch_root;
	
	/* reference element to timestamp and timevarying attribute */
	static Element ts_ref, ta_ref;
	
	static TreeWalker tw;
	
	static AnnoParser anno_parser;
	
	/* store top level elements' name and the corresponding definition in schema */
	static Hashtable def_table;
	
	static Hashtable ref_table;
	
	static Hashtable attr_table;
	
	/* parse arguments to get schema and physical annotation documents */
	static void parse_args(String[] args) throws Exception{
		for (int i = 0; i < args.length; i++){
			parser.parse(new InputSource(args[i]));
			
			Document doc = parser.getDocument();
			String root_name = doc.getDocumentElement().getNodeName();
			
			if (root_name.equals("physicalAnnotatedSchema")){
				phy_anno = doc;
				
				/* try to get non-temporal schema's location from logical annotation */
				if (schema == null){
					NodeList nl = phy_anno.getElementsByTagName("temporalAnnotatations");
					if (nl.getLength() > 0){
						String location = ((Element)nl.item(0)).getAttribute("location");
						
						parser.parse(new InputSource(location));
						Document log_anno = parser.getDocument();
						
						nl = log_anno.getElementsByTagName("snapshotSchema");
				
						if (nl.getLength() > 0){
							location = ((Element)nl.item(0)).getAttribute("location");
					
							parser.parse(new InputSource(location));
							schema = parser.getDocument();								
							
							int splash_posi = location.lastIndexOf("/");
							if (splash_posi >= 0)								
								schema_name = location.substring(splash_posi+1);
							else
								schema_name = location;									
						}
					}
				}
				
				continue;
			}
			
			if (root_name.equals("xs:schema")){
				schema_name = args[i];
				schema = doc;
			}
		}
		
		if (schema == null){
			System.out.println("Non-temporal schema is missing!");
			System.exit(0);	
		}
		
		return;		
	}
	
	public static void main(String[] args) throws Exception{		
		if (args.length == 0){			
			System.out.println("Usage: java SchemaMapper [non-temporal-schema | physical_annotation]");
			System.exit(0);
		}

		/* create reader and writer */
		parser = new DOMParser();  		
		
		System.setProperty(DOMImplementationRegistry.PROPERTY,
    		"org.apache.xerces.dom.DOMImplementationSourceImpl");
		DOMImplementationRegistry registry = DOMImplementationRegistry.newInstance();

		DOMImplementationLS impl =
			(DOMImplementationLS)registry.getDOMImplementation("LS-Load");;

		writer = impl.createDOMWriter();						
		
		/* get schema document and physical annotation document */
		parse_args(args);		
		
		if (phy_anno == null)
			create_tvroot_schema();
		else
			create_regular_schema();
	}
	
	/* create a representational schema based on non-temporal schema and physical annotation */
	static void create_regular_schema() throws Exception {
		anno_parser = new AnnoParser(null, phy_anno);
        		
		sch_root = schema.getDocumentElement();
		
		tw = ((DocumentTraversal)schema).createTreeWalker(sch_root,
													NodeFilter.SHOW_ELEMENT,
													null,
													false);
		
		/* get top level definitions */
		get_top_defs();
		
		/* 
		 *	create definition of timestamp and timevarying attribute, 
		 *	create reference to them
		 */		 	
		create_ts_def();
		create_tvattr_def();
		
		/* process time-varying attribute */
		ArrayList ta_path = anno_parser.getTimestampableTVAttrPath();
		
		for (int i = 0; i < ta_path.size(); i++){
			String path = (String)ta_path.get(i);			
		
			if (path.startsWith("/"))
				path = path.substring(1);		
			
			String[] node_name = path.split("/");
			
			/* get the definition of root element */
			Element parent = (Element)def_table.get(node_name[0]);
			
			Element next_def = parent;
			String next_name = node_name[0];			
			
			String attr_name = node_name[node_name.length - 1];
			attr_name = attr_name.substring(1);
			
			/* find the definition of elements on the path from top to bottom */
			for (int j = 1; j < node_name.length - 1; j++){
				next_name = node_name[j];
				next_def = (Element)find_node_def(parent, next_name);
				
				if (next_def.getAttribute("ref").length() > 0){					
					int ref_num = ((Integer)ref_table.get(next_name)).intValue();
					
					if (ref_num > 1){
						/*						 
						 *	change the reference declaration to embedded declaration
					 	 *	by copying the content of the refered definition here
					 	 */
					 	 
						next_def.removeAttribute("ref");
						next_def.setAttribute("name", next_name);						
						
						Element ref_def = (Element)def_table.get(next_name);
						
						NodeList content = ref_def.getChildNodes();
						
						for (int k = 0; k < content.getLength(); k++)
							next_def.appendChild(content.item(k).cloneNode(true));
							
						/* decrease the reference number of the refered definition */
						ref_table.put(next_name, new Integer(ref_num-1));
					}
					else
						/* get the refered definition */
						next_def = (Element)def_table.get(next_name);	
				}
				
				parent = next_def;
			}
			
			/* transform the definition of the attribute's owner */
			transform_ta(next_def, attr_name);
		}
		
		/* processing definitions of timestampable elements */
		ArrayList ts_path = anno_parser.getTimestampableElementsPath();				
		
		for (int i = 0; i < ts_path.size(); i++){
			String path = (String)ts_path.get(i);
			
			if (path.startsWith("/"))
				path = path.substring(1);			
			
			String[] node_name = path.split("/");
			
			/* get the defintion of root element */
			Element parent = (Element)def_table.get(node_name[0]);
			
			Element next_def = parent;
			String next_name = node_name[0];
			
			Element ref_e = null;
			
			/* find the definition of elements on the path from top to bottom */
			for (int j = 1; j < node_name.length; j++){
				next_name = node_name[j];
				next_def = find_node_def(parent, next_name);
				
				if (next_def.getAttribute("ref").length() > 0){
					int ref_num = ((Integer)ref_table.get(next_name)).intValue();
					
					if (ref_num > 1){
						/*						 
						 *	change the reference declaration to embedded declaration
					 	 *	by copying the content of the refered definition here
					 	 */

						next_def.removeAttribute("ref");
						next_def.setAttribute("name", next_name);
						
						Element ref_def = (Element)def_table.get(next_name);
						
						if (ref_def.getAttribute("type").length() > 0){
							/* element of simple type */
							next_def.setAttribute("type", ref_def.getAttribute("type"));
						}
						else{
							/* elment of complex type */
							NodeList content = ref_def.getChildNodes();
						
							for (int k = 0; k < content.getLength(); k++)
								next_def.appendChild(content.item(k).cloneNode(true));
						}
						
						/* decrease the reference number of the refered definition */
						ref_table.put(next_name, new Integer(ref_num-1));
						
						ref_e = null;
					}
					else{
						ref_e = next_def;						
						
						next_def = (Element)def_table.get(next_name);
					}
				}
				else
					ref_e = null;
					
				parent = next_def;
			}
			
			/* transform defintion of element and its parent */
			transform_element_def(next_def, ref_e);
		}
		
		transform_id_attr();
		
		/* write out temporal schema */		
		writer.writeNode(new FileOutputStream("temp_" + schema_name.substring(schema_name.lastIndexOf("/")+1)), schema);
	}	
	
	/* change ID and IDREF attribute and element's type to xs:string */
	static void transform_id_attr(){
		NodeList attr_l = sch_root.getElementsByTagName("xs:attribute");
		
		for (int i = 0; i < attr_l.getLength(); i++){
			Element attr = (Element)attr_l.item(i);
			
			if (attr.getAttribute("type").equals("xs:ID") || 
						attr.getAttribute("type").equals("xs:IDREF"))				
				attr.setAttribute("type", "xs:string");
		}	
		
		NodeList ele_l = sch_root.getElementsByTagName("xs:element");
		
		for (int i = 0; i < ele_l.getLength(); i++){
			Element ele = (Element)ele_l.item(i);
			
			if (ele.getAttribute("type").equals("xs:ID") || 
						ele.getAttribute("type").equals("xs:IDREF"))				
				ele.setAttribute("type", "xs:string");
		}	
	}
	
	/* create representational schema for TXML with time-varying root */
	static void create_tvroot_schema() throws Exception {		
		sch_root = schema.getDocumentElement();
		
		/* find the definition of root element, which is not refered in other definition */
		Hashtable top_defs = new Hashtable();
		ArrayList ref_names = new ArrayList();		
		
		NodeList cl = sch_root.getChildNodes();		
		
		/* get top-level definitions and find out those definitions that are refered by other definitons */
		for (int i = 0; i < cl.getLength(); i++){
			Node n = cl.item(i);
			
			if (!n.getNodeName().equals("xs:element"))
				continue;
			
			Element e = (Element)n;
			
			top_defs.put(e.getAttribute("name"), e);
			
			NodeList cl1 = e.getElementsByTagName("xs:element");
			for (int j = 0; j < cl1.getLength(); j++){
				Element e1 = (Element)cl1.item(j);
				
				if (e1.getAttribute("ref").length() > 0)
					ref_names.add(e1.getAttribute("ref"));
			}				
		}
		
		for (int i = 0; i < ref_names.size(); i++){						
			String name = (String)ref_names.get(i);
			
			top_defs.remove(name);
		}
		
		if (top_defs.size() != 1){			
			System.out.println("Can't decide the snapshot root among:");
			
			Enumeration enum = top_defs.elements();
			while (enum.hasMoreElements()){
				Element def = (Element)enum.nextElement();
				System.out.println(def.getAttribute("name"));
			}
			
			System.exit(0);	
		}
		
		Element root_def = (Element)top_defs.elements().nextElement();
		String root_name = root_def.getAttribute("name");		
		
		create_ts_def();
						
		/* transform the definition of root element */
		transform_element_def(root_def, null);
		
		/*
		 *	timestamps here are mandatory, not optional
		 */
		NodeList nl = root_def.getElementsByTagName("xs:element");
		for (int i = 0; i < nl.getLength(); i++){
			Element e = (Element)nl.item(i);
			
			if (e.getAttribute("ref").equals("timestamp")){
				e.removeAttribute("minOccurs");
				
				break;
			}
		}
		
		/* create definition for new timevarying root */
		Element tvroot_def = schema.createElement("xs:element");
		tvroot_def.setAttribute("name", "timevarying-root");
		
		Element type = schema.createElement("xs:complexType");
		Element comp = schema.createElement("xs:sequence");		
		
		Element item = schema.createElement("xs:element");
		item.setAttribute("ref", root_name);
		item.setAttribute("maxOccurs", "unbounded");
		
		Element verid_attr = schema.createElement("xs:attribute");
		verid_attr.setAttribute("name", "cur_verid");
		verid_attr.setAttribute("type", "xs:int");
		
		comp.appendChild(item);
		type.appendChild(comp);
		type.appendChild(verid_attr);
		tvroot_def.appendChild(type);
		
		sch_root.insertBefore(tvroot_def, sch_root.getFirstChild());
		
		transform_id_attr();
		
		writer.writeNode(new FileOutputStream("temp_" + schema_name), schema);
		
		return;
	}
	
	/* find the definition of child element within the given parent definition */
	static Element find_node_def(Element parent, String name){
		ArrayList stack = new ArrayList();
		
		tw.setCurrentNode(parent);		
		Element def = (Element)tw.firstChild();
		
		while (true){
			if (def == null){
				if (stack.size() == 0){
					System.out.println("Can't find definition of " + name + 
										" in " + parent.getAttribute("name"));	
					
					System.exit(0);
				}
				
				/* go back to parent's next sibling */
				def = (Element)stack.remove(stack.size() - 1);	
				
				tw.setCurrentNode(def);
				def = (Element)tw.nextSibling();
				
				continue;
			}
			
			if (def.getNodeName().equals("xs:element")){
				/* compare element's name */
				if (def.getAttribute("name").equals(name) ||
						def.getAttribute("ref").equals(name))
					return def;
				
				tw.setCurrentNode(def);	
				def = (Element)tw.nextSibling();
				
				continue;
			}
			else{
				/* go deep */
				stack.add(def);
				tw.setCurrentNode(def);
				def = (Element)tw.firstChild();	
			}
		}	
	}
	
	/* find the definition of attribute within the given parent definition */
	static Element find_attr_def(Element parent, String name){
		ArrayList stack = new ArrayList();
		
		tw.setCurrentNode(parent);		
		Element def = (Element)tw.firstChild();
		
		while (true){
			if (def == null){
				// go up one level
				def = (Element)stack.remove(stack.size() - 1);	
				
				tw.setCurrentNode(def);
				def = (Element)tw.nextSibling();
				
				continue;
			}
			
			String sch_name = def.getNodeName();
			
			if (sch_name.equals("xs:attribute")){
				// compare attribute name
				if (def.getAttribute("name").equals(name) ||
						def.getAttribute("ref").equals(name))
					return def;
				
				tw.setCurrentNode(def);
				def = (Element)tw.nextSibling();
				
				continue;
			}
			else if (sch_name.equals("xs:element")){				
				tw.setCurrentNode(def);
				def = (Element)tw.nextSibling();	
			}
			else {
				// go deep
				stack.add(def);
				tw.setCurrentNode(def);
				def = (Element)tw.firstChild();	
			}
		}	
	}
	
	/* 
	 *	get top-level definitions in schema and find out the reference number of elements
	 */
	static void get_top_defs(){
		def_table = new Hashtable();
		ref_table = new Hashtable();
		attr_table = new Hashtable();
    	
    	tw.setCurrentNode(sch_root);
		Element def = (Element)tw.firstChild();		
		
		// get the top level definition
		while (def != null){
			if (def.getNodeName().equals("xs:element")){
				// System.out.println("find def " + def.getAttribute("name"));
				
				// add to definition table
				def_table.put(def.getAttribute("name"), def);
				
				// find the reference in the definition of this element
				NodeList child_def = def.getElementsByTagName("xs:element");
				
				for (int i = 0; i < child_def.getLength(); i++){
					Element c = (Element)child_def.item(i);
					
					if (c.getAttribute("ref").length() == 0)
						continue;
						
					Integer ref_num = (Integer)ref_table.get(c.getAttribute("ref"));
					if (ref_num == null)
						ref_num = new Integer(1);
					else
						ref_num = new Integer(ref_num.intValue() + 1);
						
					ref_table.put(c.getAttribute("ref"), ref_num);
				}
			}
			else if (def.getNodeName().equals("xs:attribute")){
				// add to attribute table
				attr_table.put(def.getAttribute("name"), def);
			}
			
			tw.setCurrentNode(def);
			def = (Element)tw.nextSibling();
		}	
		
		return;		
	}
	
	/* create definition and reference definition of timestamp */
	static void create_ts_def(){
		Element ts_def = schema.createElement("xs:element");
		
		ts_def.setAttribute("name", "timestamp");
		
		Element type = schema.createElement("xs:complexType");		
		
		Element vtbegin = schema.createElement("xs:attribute");
		vtbegin.setAttribute("name", "vtbegin");
		vtbegin.setAttribute("type", "xs:date");
		vtbegin.setAttribute("use", "required");
		
		Element vtend = schema.createElement("xs:attribute");
		vtend.setAttribute("name", "vtend");
		vtend.setAttribute("type", "xs:date");
		vtend.setAttribute("use", "required");
		
		type.appendChild(vtbegin);
		type.appendChild(vtend);
				
		ts_def.appendChild(type);
		
		sch_root.appendChild(ts_def);
		
		ts_ref = schema.createElement("xs:element");
		ts_ref.setAttribute("ref", "timestamp");
		ts_ref.setAttribute("minOccurs", "0");		
	}
	
	/* create definition and reference definition of timevarying attribute */
	static void create_tvattr_def(){
		String name, type;		
		
		Element ta_def = schema.createElement("xs:element");
		
		ta_def.setAttribute("name", "timevarying-attr");
		
		Element sch_type = schema.createElement("xs:complexType");
		
		Element name_attr = schema.createElement("xs:attribute");
		name_attr.setAttribute("name", "name");
		name_attr.setAttribute("type", "xs:string");
		name_attr.setAttribute("use", "required");
		
		Element value_attr = schema.createElement("xs:attribute");
		value_attr.setAttribute("name", "value");
		value_attr.setAttribute("type", "xs:string");
		value_attr.setAttribute("use", "required");
		
		Element vtbegin = schema.createElement("xs:attribute");
		vtbegin.setAttribute("name", "vtbegin");
		vtbegin.setAttribute("type", "xs:date");
		vtbegin.setAttribute("use", "required");
		
		Element vtend = schema.createElement("xs:attribute");
		vtend.setAttribute("name", "vtend");
		vtend.setAttribute("type", "xs:date");
		vtend.setAttribute("use", "required");		
		
		sch_type.appendChild(name_attr);
		sch_type.appendChild(value_attr);
		sch_type.appendChild(vtbegin);
		sch_type.appendChild(vtend);		
		
		ta_def.appendChild(sch_type);
		
		sch_root.appendChild(ta_def);
		
		ta_ref = schema.createElement("xs:element");
		ta_ref.setAttribute("ref", "timevarying-attr");
		ta_ref.setAttribute("minOccurs", "0");
		ta_ref.setAttribute("maxOccurs", "unbounded");
		
		return;
	}
	
	/* 
	 *	add reference to timestamp to definiton of timestampable element,
	 *	modify its declaration in the definition of parent to allow it 
	 *	appear multiple times if it is a reference
	 */
	static void transform_element_def(Element def, Element ref){
		if (def.getAttribute("type").length() > 0){
			/*	
			 *	convert simple type to complex type and 
			 *	add reference to timestamp in compositor
			 */
			 
			def.removeAttribute("type");
			
			Element type = schema.createElement("xs:complexType");
			type.setAttribute("mixed", "true");
			
			Element comp = schema.createElement("xs:sequence");
			
			comp.appendChild(ts_ref.cloneNode(true));
			
			type.appendChild(comp);
			
			def.appendChild(type);
		}
		else{
			/* add timestamp reference to complex type */
			Element type, content;		
			
			/* get type declaration */
			Node n = def.getFirstChild();
			while (!n.getNodeName().equals("xs:complexType"))
				n = n.getNextSibling();
				
			type = (Element)n;
			
			/* get content declaration */
			n = type.getFirstChild();
			while (!(n instanceof Element))
				n = n.getNextSibling();
			
			content = (Element)n;
			
			if (content.getNodeName().equals("xs:simpleContent")){
				NodeList attr_list = content.getElementsByTagName("xs:attribute");
				
				Element new_content = schema.createElement("xs:sequence");
				new_content.appendChild(ts_ref.cloneNode(true));				
								
				for (int i = 0; i < attr_list.getLength(); i++)
					type.appendChild(attr_list.item(i).cloneNode(true));
					
				type.insertBefore(new_content, content);
				type.removeChild(content);
				
				type.setAttribute("mixed", "true");
			}
			else
				content.insertBefore(ts_ref.cloneNode(true), content.getFirstChild());
		}
		
		Element parent_compositor = null;
		
		if (ref != null){
			// relax the occur times of reference in parent
			ref.setAttribute("maxOccurs", "unbounded");
			parent_compositor = (Element)ref.getParentNode();
		}
		else{			
			// change the occur times of timestampable node
			String parent_name;
			
			parent_name = def.getParentNode().getNodeName();
			
			if (parent_name.equals("xs:all") || parent_name.equals("xs:sequence") ||
					parent_name.equals("xs:choice")){
				def.setAttribute("maxOccurs", "unbounded");
				parent_compositor = (Element)def.getParentNode();	
			}
		}
		
		/* modify parent's compositor to allow timestampable nodes to appear more than once */
		if ((parent_compositor != null) && (parent_compositor.getNodeName().equals("xs:all"))){
			Element new_compositor = schema.createElement("xs:sequence");
			
			NodeList cl = parent_compositor.getChildNodes();
			
			for (int i = 0; i < cl.getLength(); i++)
				new_compositor.appendChild(cl.item(i).cloneNode(true));
				
			Element parent_type = (Element)parent_compositor.getParentNode();
			
			parent_type.insertBefore(new_compositor, parent_compositor);
			
			parent_type.removeChild(parent_compositor);
		}
	}	
	
	static void transform_ta(Element parent, String attr_name){
		/* timevarying attribute's use in normal form is optional */
		Element attr_def = find_attr_def(parent, attr_name);
		
		/* required properity is not applicable to timestampable attribute */
		attr_def.removeAttribute("use");
		
		Element sch_type, sch_content;
		
		Node n = parent.getFirstChild();
		while (!(n instanceof Element) && !n.getNodeName().equals("xs:complexType"))
			n = n.getNextSibling();
			
		sch_type = (Element)n;
		
		// check if the parent has simple content (without child element)
		Element simple_content = null;
		
		n = sch_type.getFirstChild();
		while (n != null){
			if (n.getNodeName().equals("xs:simpleContent")){
				simple_content = (Element)n;
				break;				
			}
			
			n = n.getNextSibling();
		}
		
		if (simple_content != null){
			// change simple content to mixed xs:sequence
			Element new_content = schema.createElement("xs:sequence");
			
			new_content.appendChild(ta_ref);
			
			sch_type.insertBefore(new_content, simple_content);
			
			// copy out xs:attribute to xs:complexType
			NodeList attr_list = sch_type.getElementsByTagName("xs:attribute");
			for (int i = 0; i < attr_list.getLength(); i++)
				sch_type.insertBefore(attr_list.item(i).cloneNode(true), simple_content);
				
			sch_type.removeChild(simple_content);
			
			sch_type.setAttribute("mixed", "true");
		}
		else{
			// find compositor and insert reference to time-varying attribute to the header
			Element compositor;
			
			n = sch_type.getFirstChild();
			
			while (true){
				if (n.getNodeName().equals("xs:sequence") ||
						n.getNodeName().equals("xs:all") ||
						n.getNodeName().equals("xs:choice")){
					compositor = (Element)n;
					break;	
				}	
				
				n = n.getNextSibling();
			}
			
			if (compositor.getNodeName().equals("xs:all")){
				/* change xs:all to xs:sequence */
				Element new_compositor = schema.createElement("xs:sequence");
				
				new_compositor.appendChild(ta_ref.cloneNode(true));
			
				NodeList cl = compositor.getChildNodes();
				
				for (int i = 0; i < cl.getLength(); i++)
					new_compositor.appendChild(cl.item(i).cloneNode(true));
				
				sch_type.insertBefore(new_compositor, compositor);
				
				sch_type.removeChild(compositor);	
			}
			else if (compositor.getNodeName().equals("xs:sequence")){
				/* only one copy of reference to timevarying-attr is enough */
				n = compositor.getFirstChild();
				while (!(n instanceof Element))
					n = n.getNextSibling();
				
				/* 
				 *	timevarying attribute is processed before timevarying elements,
				 *	so reference to timevarying attribute is always the first declaration
				 *	int content
				 */	
				Element first = (Element)n;
				
				if (!first.getNodeName().equals("timevarying-attr"))
					compositor.insertBefore(ta_ref.cloneNode(true), compositor.getFirstChild());
			}
			else {
				/* 
				 *	compositor is xs:choice, create a new compositor that contains
				 *	reference to timevarying-attr and old compositor
				 */
				Element new_compositor = schema.createElement("xs:sequence");
				new_compositor.appendChild(ta_ref.cloneNode(true));
				
				new_compositor.appendChild(compositor.cloneNode(true));
				
				sch_type.insertBefore(new_compositor, compositor);
				sch_type.removeChild(compositor);
			}			
		}
	}
}
